@@ -179,7 +179,11 @@ construct_command_line(struct manager_ctx *manager, struct server *server)
     }
     if (manager->plugin) {
         int len = strlen(cmd);
-        snprintf(cmd + len, BUF_SIZE - len, " --plugin %s", manager->plugin);
+        snprintf(cmd + len, BUF_SIZE - len, " --plugin \"%s\"", manager->plugin);
+    }
+    if (manager->plugin_opts) {
+        int len = strlen(cmd);
+        snprintf(cmd + len, BUF_SIZE - len, " --plugin-opts \"%s\"", manager->plugin_opts);
     }
     for (i = 0; i < manager->nameserver_num; i++) {
         int len = strlen(cmd);
@@ -319,7 +323,7 @@ create_and_bind(const char *host, const char *port, int protocol)
 {
     struct addrinfo hints;
     struct addrinfo *result, *rp, *ipv4v6bindall;
-    int s, listen_sock, is_port_reuse;
+    int s, listen_sock = -1, is_port_reuse = 0;
 
     memset(&hints, 0, sizeof(struct addrinfo));
     hints.ai_family   = AF_UNSPEC;                  /* Return IPv4 and IPv6 choices */
@@ -356,8 +360,6 @@ create_and_bind(const char *host, const char *port, int protocol)
         }
     }
 
-    is_port_reuse = 0;
-
     for (/*rp = result*/; rp != NULL; rp = rp->ai_next) {
         listen_sock = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
         if (listen_sock == -1) {
@@ -548,7 +550,7 @@ check_port(struct manager_ctx *manager, struct server *server)
     ss_free(sock_fds);
 
     if (bind_err == -2) {
-        LOGI("port is avaiable but can not be locked");
+        LOGI("port is available but can not be locked");
     }
 
     return bind_err == -1 ? -1 : 0;
@@ -882,6 +884,7 @@ main(int argc, char **argv)
     char *iface           = NULL;
     char *manager_address = NULL;
     char *plugin          = NULL;
+    char *plugin_opts     = NULL;
 
     int auth      = 0;
     int fast_open = 0;
@@ -909,6 +912,7 @@ main(int argc, char **argv)
         { "executable",      required_argument, 0, 0 },
         { "mtu",             required_argument, 0, 0 },
         { "plugin",          required_argument, 0, 0 },
+        { "plugin-opts",     required_argument, 0, 0 },
         { "help",            no_argument,       0, 0 },
         { 0,                 0,                 0, 0 }
     };
@@ -934,6 +938,8 @@ main(int argc, char **argv)
             } else if (option_index == 5) {
                 plugin = optarg;
             } else if (option_index == 6) {
+                plugin_opts = optarg;
+            } else if (option_index == 7) {
                 usage();
                 exit(EXIT_SUCCESS);
             }
@@ -1041,9 +1047,12 @@ main(int argc, char **argv)
         if (mtu == 0) {
             mtu = conf->mtu;
         }
-        if (plugin == 0) {
+        if (plugin == NULL) {
             plugin = conf->plugin;
         }
+        if (plugin_opts == NULL) {
+            plugin_opts = conf->plugin_opts;
+        }
         if (ipv6first == 0) {
             ipv6first = conf->ipv6_first;
         }
@@ -1124,6 +1133,7 @@ main(int argc, char **argv)
     manager.nameserver_num  = nameserver_num;
     manager.mtu             = mtu;
     manager.plugin          = plugin;
+    manager.plugin_opts     = plugin_opts;
     manager.ipv6first       = ipv6first;
 #ifdef HAVE_SETRLIMIT
     manager.nofile = nofile;
