@@ -356,13 +356,13 @@ usage()
 #ifdef __linux__
     printf(
         "       [--mptcp]                  Enable Multipath TCP on MPTCP Kernel.\n");
-#ifdef MODULE_REMOTE
-    printf(
-        "       [--firewall]               Setup firewall rules for auto blocking.\n");
-#endif
 #endif
+#ifndef __MINGW32__
     printf(
-        "       [--plugin <plugin_args>]   Enable SIP003 plugin. (Experimental)\n");
+        "       [--plugin <name>]          Enable SIP003 plugin. (Experimental)\n");
+    printf(
+        "       [--plugin-opts <options>]  Set SIP003 plugin options. (Experimental)\n");
+#endif
     printf("\n");
     printf(
         "       [-v]                       Verbose mode.\n");
